library pam_flutter;

import 'dart:convert';
import 'dart:core';
import 'dart:io' show Platform;

import 'package:device_info/device_info.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart' show rootBundle;
import 'package:http/http.dart';
import 'package:package_info/package_info.dart';
import 'package:pam_flutter/pages/consent_page.dart';
import 'package:pam_flutter/pages/notification_page.dart';
import 'package:pam_flutter/pam_notification.dart';
import 'package:pam_flutter/pam_queue_manager.dart';
import 'package:pam_flutter/repositories/base_repository.dart';
import 'package:pam_flutter/repositories/consent_repository.dart';
import 'package:pam_flutter/repositories/pdpa_repository.dart';
import 'package:pam_flutter/types.dart';
import 'package:pam_flutter/utils/requester.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:uuid/uuid.dart';

import 'models/consent_message_model.dart';

class Pam {
  Pam._();

  static bool _isAppReady = false;
  static BuildContext? _buildCtx;
  static late Map<String, dynamic> _config;
  static late SharedPreferences _sharedPreferences;
  static String? _appPrefix;
  static String? _pamServer;
  static String? _publicDBAlias;
  static String? _loginDBAlias;
  static String? _baseAPI;
  static String? _trackingConsentId;
  static String? _contactingConsentId;
  static int? _refreshExpireDurationMinute;
  static late bool _enableLog;
  static late PamQueueManager _pamQueueManager;
  static List<void Function(Map<String, dynamic>)> onConsentListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onTokenListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onLaunchListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onResumeListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onMessageListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<Map<String, dynamic>> _pendingMessage =
      List<Map<String, dynamic>>.empty(growable: true);
  static late String _platformVersion;
  static late String _osVersion;
  static late String _appVersion;
  static PDPARepository? _pdpaRepository;
  static ConsentRepository? _consentRepository;

  static Future<void> initial({bool enableLog = false}) async {
    _isAppReady = false;
    _enableLog = enableLog;
    _pamQueueManager = PamQueueManager(callback: _postTracker);

    await _initialConfig();
    await _initialSharedPreferences();
    await _initialAppInfo();
    await PamNotification.pamInitial();
  }

  static Future<void> _initialConfig() async {
    try {
      String raw = await rootBundle.loadString("pam-config.json");
      _config = json.decode(raw);
    } catch (e) {
      throw (e);
    }
  }

  static Future<void> _initialSharedPreferences() async {
    _sharedPreferences = await SharedPreferences.getInstance();
  }

  static Future<void> _initialAppInfo() async {
    DeviceInfoPlugin deviceInfo = DeviceInfoPlugin();
    PackageInfo packageInfo = await PackageInfo.fromPlatform();
    _appVersion = "${packageInfo.packageName} (${packageInfo.version})";

    if (Platform.isAndroid) {
      AndroidDeviceInfo androidInfo = await deviceInfo.androidInfo;
      _osVersion = "Android: ${androidInfo.version.sdkInt}";
      _platformVersion =
          "Android: ${androidInfo.version.release} (SDK ${androidInfo.version.sdkInt}), ${androidInfo.manufacturer} ${androidInfo.model}";
    } else if (Platform.isIOS) {
      IosDeviceInfo iosInfo = await deviceInfo.iosInfo;
      _osVersion = "${iosInfo.systemName}: ${iosInfo.systemVersion}";
      _platformVersion =
          "${iosInfo.systemName}: ${iosInfo.systemVersion}, ${iosInfo.name} ${iosInfo.model}";
    }
  }

  static PDPARepository pdpaRepository(BuildContext context) {
    if (_pdpaRepository == null) {
      _pdpaRepository = PDPARepository(
        buildCtx: context,
        options: NewRepositoryOptions(
          baseUrl: "${baseAPI()}/api/pdpa/consent-message",
          findUrl: "${baseAPI()}/api/pdpa/consent-message/:id",
        ),
      );
    }
    return _pdpaRepository!;
  }

  static Future<void> refreshConsentView(
      BuildContext context, String consentId) async {
    int? expire = _getIntFromSharedPref(
        "${PamStandardSharePreferenceKey.latest_refresh_consent_status}_$consentId");
    int newExpire =
        DateTime.now().millisecondsSinceEpoch + refreshExpireDuration();
    if (expire == null) {
      _saveIntToSharedPref(
          "${PamStandardSharePreferenceKey.latest_refresh_consent_status}_$consentId",
          newExpire);
      await consentRepository(context).get(getContactId() ?? "", params: {
        "consent-message-id": consentId,
      });
    } else {
      if ((DateTime.now().millisecondsSinceEpoch - expire >
          refreshExpireDuration())) {
        _saveIntToSharedPref(
            "${PamStandardSharePreferenceKey.latest_refresh_consent_status}_$consentId",
            newExpire);
        if (getContactId() != null) {
          await consentRepository(context).get(getContactId() ?? "", params: {
            "consent-message-id": consentId,
          });
        }
      } else {
        return;
      }
    }

    if (consentRepository(context).data == null) {
      print("Consent detail is null");
    }

    if (consentRepository(context).data?.NeedConsentReview != false ||
        consentRepository(context).data == null) {
      await Navigator.of(context).push(MaterialPageRoute(builder: (context) {
        return Scaffold(
          body: ConsentPage(
            consentMessageId: consentId,
            baseAPI: baseAPI(),
            contactConsent: consentRepository(context).data,
            isSubmitTracking: true,
          ),
        );
      }));
    }
  }

  // TODO : Change pamServer() to baseAPI()
  static ConsentRepository consentRepository(BuildContext context) {
    if (_consentRepository == null) {
      _consentRepository = ConsentRepository(
        buildCtx: context,
        options: NewRepositoryOptions(
          baseUrl: "${pamServer()}/contacts",
          findUrl: "${pamServer()}/contacts/:id/consents",
        ),
      );
    }
    return _consentRepository!;
  }

  static Future<ConsentMessageSettingModel?> consentRequestView(
    BuildContext context,
    String consentId, {
    bool isSubmitTracking = true,
  }) async {
    if (getContactId() != null) {
      await consentRepository(context).get(getContactId() ?? "", params: {
        "consent-message-id": consentId,
      });
    }

    if (consentRepository(context).data == null) {
      print("Consent detail is null");
    }

    return await Navigator.of(context)
        .push(MaterialPageRoute(builder: (context) {
      return Scaffold(
        body: ConsentPage(
          consentMessageId: consentId,
          baseAPI: baseAPI(),
          contactConsent: consentRepository(context).data,
          isSubmitTracking: isSubmitTracking,
        ),
      );
    }));
  }

  static Future<dynamic> notificationView(BuildContext context) async {
    return await Navigator.of(context)
        .push(MaterialPageRoute(builder: (context) {
      return Scaffold(
        body: NotificationsPage(
          baseAPI: baseAPI(),
        ),
      );
    }));
  }

  static Future<Widget?> consentRequestWidget(
      BuildContext context, String consentId) async {
    if (getContactId() == null) {
      print("Contact is required");
      return null;
    } else {
      await consentRepository(context).get(getContactId()!, params: {
        "consent-message-id": consentId,
      });

      if (consentRepository(context).data == null) {
        print("Consent detail is null");
        return null;
      } else {
        return ConsentPage(
          consentMessageId: consentId,
          baseAPI: baseAPI(),
          contactConsent: consentRepository(context).data!,
          isSubmitTracking: true,
        );
      }
    }
  }

  static Widget notificationWidget() {
    return NotificationsPage(
      baseAPI: baseAPI(),
    );
  }

  static String appPrefix() {
    if (_appPrefix == null) {
      _appPrefix = _config["app-prefix"];
    }
    return _appPrefix!;
  }

  static String pamServer() {
    if (_pamServer == null) {
      _pamServer = _config["pam-server"];
    }
    return _pamServer!;
  }

  static String publicDBAlias() {
    if (_publicDBAlias == null) {
      _publicDBAlias = _config["public-db-alias"];
    }
    return _publicDBAlias!;
  }

  static String loginDBAlias() {
    if (_loginDBAlias == null) {
      _loginDBAlias = _config["login-db-alias"];
    }
    return _loginDBAlias!;
  }

  static String baseAPI() {
    if (_baseAPI == null) {
      _baseAPI = _config["base-api"];
    }
    return _baseAPI!;
  }

  static String? trackingConsentId() {
    if (_trackingConsentId == null) {
      _trackingConsentId = _config["tracking-consent-id"];
    }
    return _trackingConsentId;
  }

  static String? contactingConsentId() {
    if (_contactingConsentId == null) {
      _contactingConsentId = _config["contacting-consent-id"];
    }
    return _contactingConsentId;
  }

  static int refreshExpireDuration() {
    if (_refreshExpireDurationMinute == null) {
      int? minute = int.parse(_config["refresh_expire_duration_minute"] ?? 5);
      _refreshExpireDurationMinute = 1000 * 60 * (minute);
    }
    return _refreshExpireDurationMinute!;
  }

  static void listen(
      String eventName, void Function(Map<String, dynamic>) callback) {
    if (eventName.toLowerCase() == PamStandardCallback.on_token) {
      onTokenListener.add(callback);
    } else if (eventName.toLowerCase() == PamStandardCallback.on_launch) {
      onLaunchListener.add(callback);
    } else if (eventName.toLowerCase() == PamStandardCallback.on_resume) {
      onResumeListener.add(callback);
    } else if (eventName.toLowerCase() == PamStandardCallback.on_message) {
      onMessageListener.add(callback);
    }
  }

  static void dispatch(String eventName, Map<String, dynamic> payload) {
    if (eventName.toLowerCase() == PamStandardCallback.on_consent) {
      onConsentListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_token) {
      onTokenListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_launch) {
      onLaunchListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_resume) {
      onResumeListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_message) {
      onMessageListener.forEach((cb) {
        cb.call(payload);
      });
    }
  }

  static Future<void> appReady(BuildContext context) async {
    if (!_isAppReady) {
      _isAppReady = true;
      _buildCtx = context;
      _track(PamStandardEvent.app_launch);
      _pendingMessage.forEach((payload) {
        dispatch(PamStandardCallback.on_message, payload);
      });
      _pendingMessage.clear();
    }
  }

  static Future<void> askNotificationPermission() async {
    return PamNotification.askNotificationPermission();
  }

  static dynamic _getFromSharedPref(String key) {
    return _sharedPreferences.get(key);
  }

  static String? _getStringFromSharedPref(String key) {
    return _sharedPreferences.getString(key);
  }

  static int? _getIntFromSharedPref(String key) {
    return _sharedPreferences.getInt(key);
  }

  static Future<bool> _saveStringToSharedPref(String key, String value) {
    return _sharedPreferences.setString(key, value);
  }

  static Future<bool> _saveIntToSharedPref(String key, int value) {
    return _sharedPreferences.setInt(key, value);
  }

  static Future<bool> _removeFromSharedPref(String key) {
    return _sharedPreferences.remove(key);
  }

  static Future<String> _getSessionId() async {
    int? expire = _getIntFromSharedPref(
        PamStandardSharePreferenceKey.session_expire_in_mils);
    String? session =
        _getStringFromSharedPref(PamStandardSharePreferenceKey.session_id);

    if ((expire == null || session == null) ||
        (DateTime.now().millisecondsSinceEpoch - expire > (1000 * 60 * 60))) {
      int newExpire = DateTime.now().millisecondsSinceEpoch + 1000 * 60 * 60;
      session = Uuid().v4();

      await _saveIntToSharedPref(
          PamStandardSharePreferenceKey.session_expire_in_mils, newExpire);
      await _saveStringToSharedPref(
          PamStandardSharePreferenceKey.session_id, session);
    } else {
      int newExpire = DateTime.now().millisecondsSinceEpoch + 1000 * 60 * 60;

      await _saveIntToSharedPref(
          PamStandardSharePreferenceKey.session_expire_in_mils, newExpire);
      await _saveStringToSharedPref(
          PamStandardSharePreferenceKey.session_id, session);
    }

    return session;
  }

  static String? getContactId() {
    if (_isUserLogin()) {
      if (_getStringFromSharedPref(
              PamStandardSharePreferenceKey.login_contact_id) !=
          null) {
        return _getStringFromSharedPref(
            PamStandardSharePreferenceKey.login_contact_id);
      } else {
        return _getStringFromSharedPref(
            PamStandardSharePreferenceKey.contact_id);
      }
    } else {
      return _getStringFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    }
  }

  static String? _getCustomerId() {
    return _getStringFromSharedPref(PamStandardSharePreferenceKey.customer_id);
  }

  static Future<void> userLogin(String customerId) async {
    await _saveStringToSharedPref(
        PamStandardSharePreferenceKey.customer_id, customerId);
    _track(
      PamStandardEvent.login,
      payload: {
        "customer": customerId,
      },
    );
  }

  static Future<void> userLogout() async {
    Map<String, dynamic> payload = {};
    if (Platform.isAndroid) {
      payload["_delete_media"] = {"android_notification": ""};
    } else if (Platform.isIOS) {
      payload["_delete_media"] = {"ios_notification": ""};
    }

    _track(
      PamStandardEvent.logout,
      payload: payload,
      deleteLoginContactAfterPost: true,
    );

    await _removeFromSharedPref(PamStandardSharePreferenceKey.customer_id);
    await _removeFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    await _removeFromSharedPref(PamStandardSharePreferenceKey.login_contact_id);
  }

  static Future<void> trackPageView({
    required String url,
    required String title,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.page_view, payload: {
      "page_url": url,
      "page_title": title,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackFavourite({
    required String id,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.favourite, payload: {
      "product_id": id,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackUnFavourite({
    required String id,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.remove_favourite, payload: {
      "product_id": id,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackAddToCart({
    required String id,
    String? title,
    double? price,
    int? quantity,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.add_to_cart, payload: {
      "page_title": title,
      "product_id": id,
      "product_title": title,
      "product_price": price,
      "quantity": quantity ?? 1,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackRemoveFromCart({
    required String id,
    String? title,
    double? price,
    int? quantity,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.remove_from_cart, payload: {
      "page_title": title,
      "product_id": id,
      "product_title": title,
      "product_price": price,
      "quantity": quantity ?? 1,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackPurchaseSuccess({
    required List<String> ids,
    required List<String> titles,
    required List<String> categories,
    required double? totalPrice,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.purchase_success, payload: {
      "product_id": ids.join(","),
      "product_title": titles.join(","),
      "product_cat": categories.join(","),
      "total_price": totalPrice,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackSavePush({
    required String token,
    Map<String, dynamic>? others,
  }) async {
    Map<String, dynamic> pl = {};
    if (Platform.isAndroid) {
      pl["android_notification"] = token;
    } else if (Platform.isIOS) {
      pl["ios_notification"] = token;
    }

    dispatch(PamStandardCallback.on_token, {
      "token": token,
    });

    return _track(PamStandardEvent.save_push, payload: {
      ...pl,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackCustomEvent({
    required String eventName,
    Map<String, dynamic>? payload,
  }) async {
    return _track(PamStandardEvent.page_view, payload: {
      ...(payload ?? Map<String, dynamic>.from({})),
    });
  }

  // Use this function to ensure that track must finish before proceed
  static Future<IPamResponse?> asyncTrackCustomEvent({
    required String eventName,
    required Map<String, dynamic> payload,
    bool deleteLoginContactAfterPost = false,
  }) async {
    if (getContactId() == null && eventName != PamStandardEvent.allow_consent) {
      return null;
    } else if (_buildCtx != null) {
      if (consentRepository(_buildCtx!).data?.NeedConsentReview == false) {
        return null;
      }
    } else if (_buildCtx != null) {
      throw ("You have to call appReady()");
    }

    String url = "${pamServer()}/trackers/events";
    Map<String, dynamic> body = {
      "event": eventName,
      "platform": _platformVersion,
    };
    Map<String, dynamic> formField = {
      "os_version": _osVersion,
      "app_version": _appVersion,
      "_session_id": await _getSessionId(),
      "_consent_message_id": trackingConsentId(),
    };

    if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.login_contact_id) !=
        null) {
      formField["_contact_id"] = _getStringFromSharedPref(
          PamStandardSharePreferenceKey.login_contact_id);
    } else if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.contact_id) !=
        null) {
      formField["_contact_id"] =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    }

    payload.forEach((key, value) {
      if (key == "page_url" || key == "page_title") {
        body[key] = value;
      } else {
        formField[key] = value;
      }
    });

    if (_isUserLogin()) {
      formField["_database"] = loginDBAlias();
      if (_getCustomerId() != null) {
        formField["customer"] = _getCustomerId();
      }
    } else {
      formField["_database"] = publicDBAlias();
    }

    body["form_fields"] = formField;

    if (_enableLog) {
      print("🦄 : POST Event = 🍀$eventName🍀");
      print("🦄 : Payload");
      print(
          "🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n$body\n🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n\n");
    }
    IPamResponse? pamResponse;
    Response response = await Requester.post(url, body);

    if (_enableLog) {
      print("Track response is ${response.body}\n");
    }

    if (response.statusCode == 200) {
      pamResponse = IPamResponse.fromJson(json.decode(response.body));
      if (pamResponse.ContactId != null && pamResponse.ContactId != "") {
        if (_isUserLogin()) {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.login_contact_id,
              pamResponse.ContactId!);
        } else {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.contact_id, pamResponse.ContactId!);
        }
      }
      if (pamResponse.ConsentId != null && pamResponse.ConsentId != "") {
        Pam.dispatch(PamStandardCallback.on_consent, {
          "consent_id": pamResponse.ConsentId ?? "",
        });
      }
    } else {
      if (_enableLog) {
        print("Track error ${response.body}\n");
      }
    }

    if (deleteLoginContactAfterPost) {
      _removeFromSharedPref(PamStandardSharePreferenceKey.customer_id);
      _removeFromSharedPref(PamStandardSharePreferenceKey.login_contact_id);
    }

    if (_enableLog) {
      print("Track has been send\n");
    }

    return pamResponse;
  }

  static void _track(
    String eventName, {
    Map<String, dynamic>? payload,
    bool deleteLoginContactAfterPost = false,
  }) {
    return _pamQueueManager.enqueue(eventName,
        payload: payload,
        deleteLoginContactAfterPost: deleteLoginContactAfterPost);
  }

  static bool _isUserLogin() {
    return _getStringFromSharedPref(
            PamStandardSharePreferenceKey.customer_id) !=
        null;
  }

  static Future<void> _postTracker(
    String eventName,
    Map<String, dynamic> payload, {
    bool deleteLoginContactAfterPost = false,
  }) async {
    print("1");
    if (getContactId() == null && eventName != PamStandardEvent.allow_consent) {
      print("2");
      return;
    } else if (_buildCtx != null) {
      print("3");
      if (consentRepository(_buildCtx!).data?.NeedConsentReview == false) {
        print("4");
        return;
      }
    } else if (_buildCtx == null) {
      print("5");
      throw ("You have to call appReady()");
    }
    print("6");

    String url = "${pamServer()}/trackers/events";
    Map<String, dynamic> body = {
      "event": eventName,
      "platform": _platformVersion,
    };
    Map<String, dynamic> formField = {
      "os_version": _osVersion,
      "app_version": _appVersion,
      "_session_id": await _getSessionId(),
      "_consent_message_id": trackingConsentId(),
    };

    if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.login_contact_id) !=
        null) {
      formField["_contact_id"] = _getStringFromSharedPref(
          PamStandardSharePreferenceKey.login_contact_id);
    } else if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.contact_id) !=
        null) {
      formField["_contact_id"] =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    }

    payload.forEach((key, value) {
      if (key == "page_url" || key == "page_title") {
        body[key] = value;
      } else {
        formField[key] = value;
      }
    });

    if (_isUserLogin()) {
      formField["_database"] = loginDBAlias();
      if (_getCustomerId() != null) {
        formField["customer"] = _getCustomerId();
      }
    } else {
      formField["_database"] = publicDBAlias();
    }

    body["form_fields"] = formField;

    if (_enableLog) {
      print("🦄 : POST Event = 🍀$eventName🍀");
      print("🦄 : Payload");
      print(
          "🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n$body\n🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n\n");
    }

    Response response = await Requester.post(url, body);
    if (_enableLog) {
      print("Track response is ${response.body}\n");
    }

    if (response.statusCode == 200) {
      IPamResponse pamResponse =
          IPamResponse.fromJson(json.decode(response.body));
      if (pamResponse.ContactId != null && pamResponse.ContactId != "") {
        if (_isUserLogin()) {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.login_contact_id,
              pamResponse.ContactId!);
        } else {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.contact_id, pamResponse.ContactId!);
        }
      }
      if (pamResponse.ConsentId != null && pamResponse.ConsentId != "") {
        Pam.dispatch(PamStandardCallback.on_consent, {
          "consent_id": pamResponse.ConsentId ?? "",
        });
      }
    } else {
      if (_enableLog) {
        print("Track error ${response.body}\n");
      }
    }

    if (deleteLoginContactAfterPost) {
      _removeFromSharedPref(PamStandardSharePreferenceKey.customer_id);
      _removeFromSharedPref(PamStandardSharePreferenceKey.login_contact_id);
    }
    _pamQueueManager.nextQueue();

    if (_enableLog) {
      print("Track has been send\n");
    }
  }
}
